@@ -2,7 +2,6 @@
 
 require 'rubygems'
 require 'rake'
-require 'rake/contrib/sshpublisher'
 require 'rake/gempackagetask'
 require 'rake/rdoctask'
 require 'rake/testtask'
@@ -10,6 +9,11 @@ require 'rbconfig'
 require 'rubyforge'
 require 'yaml'
 
+begin
+  gem 'rdoc'
+rescue Gem::LoadError
+end
+
 ##
 # hoe - a tool to help rake
 #
@@ -34,27 +38,31 @@ require 'yaml'
 #
 # === Tasks Provided:
 #
-# announce::       Create news email file and post to rubyforge.
-# audit::          Run ZenTest against the package.
-# check_manifest:: Verify the manifest.
-# clean::          Clean up all the extras.
-# config_hoe::     Create a fresh ~/.hoerc file.
-# debug_gem::      Show information about the gem.
-# default::        Run the default tasks.
-# docs::           Build the docs HTML Files
-# email::          Generate email announcement file.
-# gem::            Build the gem file hoe-1.5.0.gem
-# generate_key::   Generate a key for signing your gems.
-# install_gem::    Install the package as a gem.
-# multi::          Run the test suite using multiruby.
-# package::        Build all the packages
-# post_blog::      Post announcement to blog.
-# post_news::      Post announcement to rubyforge.
-# publish_docs::   Publish RDoc to RubyForge.
-# release::        Package and upload the release to rubyforge.
-# ridocs::         Generate ri locally for testing.
-# test::           Run the test suite.
-# test_deps::      Show which test files fail when run alone.
+# announce::          Create news email file and post to rubyforge.
+# audit::             Run ZenTest against the package.
+# check_manifest::    Verify the manifest.
+# clean::             Clean up all the extras.
+# config_hoe::        Create a fresh ~/.hoerc file.
+# debug_gem::         Show information about the gem.
+# default::           Run the default tasks.
+# deps:email::        Print a contact list for gems dependent on this gem
+# deps:fetch::        Fetch all the dependent gems of this gem into tarballs
+# deps:list::         List all the dependent gems of this gem
+# docs::              Build the docs HTML Files
+# email::             Generate email announcement file.
+# gem::               Build the gem file hoe-1.8.0.gem
+# generate_key::      Generate a key for signing your gems.
+# install_gem::       Install the package as a gem.
+# multi::             Run the test suite using multiruby.
+# package::           Build all the packages
+# post_blog::         Post announcement to blog.
+# post_news::         Post announcement to rubyforge.
+# publish_docs::      Publish RDoc to RubyForge.
+# release::           Package and upload the release to rubyforge.
+# ridocs::            Generate ri locally for testing.
+# tasks::             Generate a list of tasks for doco.
+# test::              Run the test suite.
+# test_deps::         Show which test files fail when run alone.
 #
 # === Extra Configuration Options:
 #
@@ -64,8 +72,8 @@ require 'yaml'
 # exclude::             A regular expression of files to exclude from
 #                       +check_manifest+.
 # publish_on_announce:: Run +publish_docs+ when you run +release+.
-# signing_key_file:: Signs your gems with this private key.
-# signing_cert_file:: Signs your gem with this certificate.
+# signing_key_file::    Signs your gems with this private key.
+# signing_cert_file::   Signs your gem with this certificate.
 # blogs::               An array of hashes of blog settings.
 #
 # Run +config_hoe+ and see ~/.hoerc for examples.
@@ -116,7 +124,8 @@ require 'yaml'
 #
 
 class Hoe
-  VERSION = '1.5.1'
+  VERSION = '1.8.3'
+  GEMURL = URI.parse 'http://gems.rubyforge.org' # for namespace :deps below
 
   ruby_prefix = Config::CONFIG['prefix']
   sitelibdir = Config::CONFIG['sitelibdir']
@@ -154,7 +163,7 @@ class Hoe
 
   DLEXT = Config::CONFIG['DLEXT']
 
-  WINDOZE = /djgpp|(cyg|ms|bcc)win|mingw/ =~ RUBY_PLATFORM unless defined? WINDOZE
+  WINDOZE = /mswin|mingw/ =~ RUBY_PLATFORM unless defined? WINDOZE
 
   DIFF = if WINDOZE
            'diff.exe'
@@ -179,6 +188,11 @@ class Hoe
 
   attr_accessor :bin_files # :nodoc:
 
+  ##
+  # *Optional*: An array of the project's blog categories. Defaults to project name.
+
+  attr_accessor :blog_categories
+
   ##
   # Optional: A description of the release's latest changes. Auto-populates.
 
@@ -209,11 +223,21 @@ class Hoe
 
   attr_accessor :extra_deps
 
+  ##
+  # Optional: An array of rubygem developer dependencies.
+
+  attr_accessor :extra_dev_deps
+
   ##
   # Populated automatically from the manifest. List of library files.
 
   attr_accessor :lib_files # :nodoc:
 
+  ##
+  # Optional: Array of incompatible versions for multiruby filtering. Used as a regex.
+
+  attr_accessor :multiruby_skip
+
   ##
   # *MANDATORY*: The name of the release.
 
@@ -229,6 +253,11 @@ class Hoe
 
   attr_accessor :need_zip
 
+  ##
+  # Optional: A post-install message to be displayed when gem is installed.
+
+  attr_accessor :post_install_message
+
   ##
   # Optional: A regexp to match documentation files against the manifest.
 
@@ -279,6 +308,11 @@ class Hoe
 
   attr_accessor :test_globs
 
+  ##
+  # Optional: What test library to require [default: test/unit]
+
+  attr_accessor :testlib
+
   ##
   # Optional: The url(s) of the project. (can be array). Auto-populates.
 
@@ -289,17 +323,47 @@ class Hoe
 
   attr_accessor :version
 
+  ##
+  # Add extra dirs to both $: and RUBY_FLAGS (for test runs)
+
+  def self.add_include_dirs(*dirs)
+    dirs = dirs.flatten
+    $:.unshift(*dirs)
+    s = File::PATH_SEPARATOR
+    Hoe::RUBY_FLAGS.sub!(/-I/, "-I#{dirs.join(s)}#{s}")
+  end
+
+  def self.normalize_names project
+    project   = project.tr('-', '_').gsub(/([A-Z])/, '_\1').downcase.sub(/^_/, '')
+    klass     = project.gsub(/(?:^|_)([a-z])/) { $1.upcase }
+    file_name = project
+
+    return project, file_name, klass
+  end
+
+  def normalize_deps deps
+    Array(deps).map { |o| String === o ? [o] : o }
+  end
+
+  def missing name
+    warn "** #{name} is missing or in the wrong format for auto-intuiting."
+    warn "   run `sow blah` and look at its text files"
+  end
+
   def initialize(name, version) # :nodoc:
     self.name = name
     self.version = version
 
     # Defaults
     self.author = []
-    self.clean_globs = %w(diff diff.txt email.txt ri
+    self.clean_globs = %w(diff diff.txt email.txt ri deps .source_index
                           *.gem *~ **/*~ *.rbc **/*.rbc)
     self.description_sections = %w(description)
+    self.blog_categories = [name]
     self.email = []
     self.extra_deps = []
+    self.extra_dev_deps = []
+    self.multiruby_skip = []
     self.need_tar = true
     self.need_zip = false
     self.rdoc_pattern = /^(lib|bin|ext)|txt$/
@@ -309,12 +373,14 @@ class Hoe
     self.spec_extras = {}
     self.summary_sentences = 1
     self.test_globs = ['test/**/test_*.rb']
+    self.testlib = 'test/unit'
+    self.post_install_message = nil
 
     yield self if block_given?
 
     # Intuit values:
 
-    readme   = File.read("README.txt").split(/^(=+ .*)$/)[1..-1]
+    readme   = File.read("README.txt").split(/^(=+ .*)$/)[1..-1] rescue ''
     unless readme.empty? then
       sections = readme.map { |s|
         s =~ /^=/ ? s.strip.downcase.chomp(':').split.last : s.strip
@@ -324,14 +390,20 @@ class Hoe
       summ = desc.split(/\.\s+/).first(summary_sentences).join(". ")
 
       self.description ||= desc
-      self.changes ||= File.read("History.txt").split(/^(===.*)/)[1..2].join.strip
       self.summary ||= summ
       self.url ||= readme[1].gsub(/^\* /, '').split(/\n/).grep(/\S+/)
     else
-      warn "** README.txt is in the wrong format for auto-intuiting."
-      warn "   run sow blah and look at it's text files"
+      missing 'README.txt'
     end
 
+    self.changes ||= begin
+                       h = File.read("History.txt")
+                       h.split(/^(===.*)/)[1..2].join.strip
+                     rescue
+                       missing 'History.txt'
+                       ''
+                     end
+
     %w(email author).each do |field|
       value = self.send(field)
       if value.nil? or value.empty? then
@@ -339,7 +411,7 @@ class Hoe
           warn "Hoe #{field} value not set - Fix by 2008-04-01!"
           self.send "#{field}=", "doofus"
         else
-          abort "Hoe #{field} value not set"
+          abort "Hoe #{field} value not set. aborting"
         end
       end
     end
@@ -349,14 +421,15 @@ class Hoe
       'rubyforge' => ">= #{::RubyForge::VERSION}",
     }
 
-    self.extra_deps = Array(extra_deps).map { |o| String === o ? [o] : o }
+    self.extra_deps     = normalize_deps extra_deps
+    self.extra_dev_deps = normalize_deps extra_dev_deps
 
     if name == 'hoe' then
       hoe_deps.each do |pkg, vers|
         extra_deps << [pkg, vers]
       end
     else
-      extra_deps << ['hoe', ">= #{VERSION}"] unless hoe_deps.has_key? name
+      extra_dev_deps << ['hoe', ">= #{VERSION}"] unless hoe_deps.has_key? name
     end
 
     define_tasks
@@ -367,14 +440,14 @@ class Hoe
     self.email << email
   end
 
-  def define_tasks # :nodoc:
-    def with_config # :nodoc:
-      rc = File.expand_path("~/.hoerc")
-      exists = File.exist? rc
-      config = exists ? YAML.load_file(rc) : {}
-      yield(config, rc)
-    end
+  def with_config # :nodoc:
+    rc = File.expand_path("~/.hoerc")
+    exists = File.exist? rc
+    config = exists ? YAML.load_file(rc) : {}
+    yield(config, rc)
+  end
 
+  def define_tasks # :nodoc:
     desc 'Run the default tasks.'
     task :default => :test
 
@@ -387,8 +460,11 @@ class Hoe
     task :test_deps do
       tests = Dir["test/**/test_*.rb"]  +  Dir["test/**/*_test.rb"]
 
+      paths = ['bin', 'lib', 'test'].join(File::PATH_SEPARATOR)
+      null_dev = WINDOZE ? '> NUL 2>&1' : '&> /dev/null'
+
       tests.each do |test|
-        if not system "ruby -Ibin:lib:test #{test} &> /dev/null" then
+        if not system "ruby -I#{paths} #{test} #{null_dev}" then
           puts "Dependency Issues: #{test}"
         end
       end
@@ -434,6 +510,10 @@ class Hoe
         s.add_dependency(*dep)
       end
 
+      extra_dev_deps.each do |dep|
+        s.add_development_dependency(*dep)
+      end
+
       s.files = File.read("Manifest.txt").delete("\r").split(/\n/)
       s.executables = s.files.grep(/^bin/) { |f| File.basename(f) }
 
@@ -445,6 +525,8 @@ class Hoe
       s.extra_rdoc_files = s.files.grep(/txt$/)
       s.has_rdoc = true
 
+      s.post_install_message = post_install_message
+
       if test ?f, "test/test_all.rb" then
         s.test_file = "test/test_all.rb"
       else
@@ -460,29 +542,32 @@ class Hoe
       # Allow automatic inclusion of compiled extensions
       if ENV['INLINE'] then
         s.platform = ENV['FORCE_PLATFORM'] || Gem::Platform::CURRENT
-        # name of the extension is CamelCase
-        if name =~ /[A-Z]/
-          # ClassName => class_name
-          alternate_name = name.reverse.scan(%r/[A-Z]+|[^A-Z]*[A-Z]+?/).reverse.map { |word| word.reverse.downcase }.join('_')
-        elsif name =~ /_/
-          # class_name = ClassName
-          alternate_name = name.strip.split(/\s*_+\s*/).map! { |w| w.downcase.sub(/^./) { |c| c.upcase } }.join
-        end
 
         # Try collecting Inline extensions for +name+
         if defined?(Inline) then
           directory 'lib/inline'
 
-          extensions = Dir.chdir(Inline::directory) {
-            Dir["Inline_{#{name},#{alternate_name}}_*.#{DLEXT}"]
-          }
-          extensions.each do |ext|
-            # add the inlined extension to the spec files
-            s.files += ["lib/inline/#{ext}"]
+          Inline.registered_inline_classes.each do |cls|
+            name = cls.name # TODO: what about X::Y::Z?
+            # name of the extension is CamelCase
+            alternate_name = if name =~ /[A-Z]/ then
+                               name.gsub(/([A-Z])/, '_\1').downcase.sub(/^_/, '')
+                             elsif name =~ /_/ then
+                               name.capitalize.gsub(/_([a-z])/) { $1.upcase }
+                             end
 
-            # include the file in the tasks
-            file "lib/inline/#{ext}" => ["lib/inline"] do
-              cp File.join(Inline::directory, ext), "lib/inline"
+            extensions = Dir.chdir(Inline::directory) {
+              Dir["Inline_{#{name},#{alternate_name}}_*.#{DLEXT}"]
+            }
+
+            extensions.each do |ext|
+              # add the inlined extension to the spec files
+              s.files += ["lib/inline/#{ext}"]
+
+              # include the file in the tasks
+              file "lib/inline/#{ext}" => ["lib/inline"] do
+                cp File.join(Inline::directory, ext), "lib/inline"
+              end
             end
           end
         end
@@ -515,7 +600,8 @@ class Hoe
 
     desc 'Install the package as a gem.'
     task :install_gem => [:clean, :package] do
-      sh "#{'sudo ' unless WINDOZE}gem install --local pkg/*.gem"
+      gem = Dir['pkg/*.gem'].first
+      sh "#{'sudo ' unless WINDOZE}gem install --local #{gem}"
     end
 
     desc 'Package and upload the release to rubyforge.'
@@ -529,7 +615,7 @@ class Hoe
         puts "rf.add_file #{rubyforge_name.inspect}, #{name.inspect}, release_id, \"#{pkg}.gem\""
       end
 
-      rf = RubyForge.new
+      rf = RubyForge.new.configure
       puts "Logging in"
       rf.login
 
@@ -551,7 +637,8 @@ class Hoe
 
     Rake::RDocTask.new(:docs) do |rd|
       rd.main = "README.txt"
-      rd.options << '-d' if RUBY_PLATFORM !~ /win32/ and `which dot` =~ /\/dot/ and not ENV['NODOT']
+      rd.options << '-d' if
+        `which dot` =~ /\/dot/ unless ENV['NODOT'] unless WINDOZE
       rd.rdoc_dir = 'doc'
       files = spec.files.grep(rdoc_pattern)
       files -= ['Manifest.txt']
@@ -586,6 +673,133 @@ class Hoe
       end
     end
 
+    ############################################################
+    # Dependencies:
+
+    namespace :deps do
+      require 'zlib' # HACK for rubygems 1.3.0
+      require 'rubygems/remote_fetcher'
+
+      @@index = nil
+
+      def self.get_source_index
+        return @@index if @@index
+
+        dump = unless File.exist? '.source_index' then
+                 url = GEMURL + "Marshal.#{Gem.marshal_version}.Z"
+                 dump = Gem::RemoteFetcher.fetcher.fetch_path url
+                 dump = Gem.inflate dump
+                 open '.source_index', 'wb' do |io| io.write dump end
+                 dump
+               else
+                 open '.source_index', 'rb' do |io| io.read end
+               end
+
+        @@index = Marshal.load dump
+      end
+
+      def self.get_latest_gems
+        @@cache ||= get_source_index.latest_specs
+      end
+
+      def self.get_gems_by_name
+        @@by_name ||= Hash[*get_latest_gems.map { |gem|
+                             [gem.name, gem, gem.full_name, gem]
+                           }.flatten]
+      end
+
+      def self.dependent_upon name
+        get_latest_gems.find_all { |gem|
+          gem.dependencies.any? { |dep| dep.name == name }
+        }
+      end
+
+
+      desc "List all the dependent gems of this gem"
+      task :list do
+        gems = self.get_gems_by_name
+        gem  = gems[self.name]
+
+        abort "Couldn't find gem: #{self.name}" unless gem
+
+        deps = self.dependent_upon self.name
+        max  = deps.map { |s| s.full_name.size }.max
+
+        puts "  dependents:"
+        unless deps.empty? then
+          deps.sort_by { |spec| spec.full_name }.each do |spec|
+            vers = spec.dependencies.find {|s| s.name == name }.requirement_list
+            puts "    %-*s - %s" % [max, spec.full_name, vers.join(", ")]
+          end
+        else
+          puts "    none"
+        end
+      end
+
+      desc "Print a contact list for gems dependent on this gem"
+      task :email do
+        gems = self.get_gems_by_name
+        gem  = gems[self.name]
+
+        abort "Couldn't find gem: #{self.name}" unless gem
+
+        deps = self.dependent_upon self.name
+
+        email = deps.map { |s| s.email }.flatten.sort.uniq
+        email = email.map { |s| s.split(/,\s*/) }.flatten.sort.uniq
+
+        email.map! { |s| # don't you people realize how easy this is?
+          s.gsub(/ at | _at_ |\s*(atmark|@nospam@|-at?-|@at?@|<at?>|\[at?\]|\(at?\))\s*/i, '@').gsub(/\s*(dot|\[d(ot)?\]|\.dot\.)\s*/i, '.').gsub(/\s+com$/, '.com')
+        }
+
+        bad, good = email.partition { |e| e !~ /^[\w.+-]+\@[\w.+-]+$/ }
+
+        warn "Rejecting #{bad.size} email. I couldn't unmunge them." unless
+          bad.empty?
+
+        puts good.join(", ")
+      end
+
+      desc "Fetch all the dependent gems of this gem into tarballs"
+      task :fetch do
+        gems = self.get_gems_by_name
+        gem  = gems[self.name]
+        deps = self.dependent_upon self.name
+
+        mkdir "deps" unless File.directory? "deps"
+        Dir.chdir "deps" do
+          begin
+            deps.sort_by { |spec| spec.full_name }.each do |spec|
+              full_name = spec.full_name
+              tgz_name  = "#{full_name}.tgz"
+              gem_name  = "#{full_name}.gem"
+
+              next if File.exist? tgz_name
+              FileUtils.rm_rf [full_name, gem_name]
+
+              begin
+                warn "downloading #{full_name}"
+                Gem::RemoteFetcher.fetcher.download(spec, GEMURL, Dir.pwd)
+                FileUtils.mv "cache/#{gem_name}", '.'
+              rescue Gem::RemoteFetcher::FetchError
+                warn "  failed"
+                next
+              end
+
+              warn "converting #{gem_name} to tarball"
+
+              system "gem unpack #{gem_name} 2> /dev/null"
+              system "gem spec -l #{gem_name} > #{full_name}/gemspec.rb"
+              system "tar zmcf #{tgz_name} #{full_name}"
+              FileUtils.rm_rf [full_name, gem_name, "cache"]
+            end
+          ensure
+            FileUtils.rm_rf "cache"
+          end
+        end
+      end
+    end
+
     ############################################################
     # Misc/Maintenance:
 
@@ -599,7 +813,7 @@ class Hoe
     task :clean => [ :clobber_docs, :clobber_package ] do
       clean_globs.each do |pattern|
         files = Dir[pattern]
-        rm_rf files unless files.empty?
+        rm_rf files, :verbose => true unless files.empty?
       end
     end
 
@@ -662,7 +876,9 @@ class Hoe
         config['blogs'].each do |site|
           server = XMLRPC::Client.new2(site['url'])
           content = site['extra_headers'].merge(:title => title,
-                                                :description => body)
+                                                :description => body,
+                                                :categories => blog_categories)
+
           result = server.call('metaWeblog.newPost',
                                site['blog_id'],
                                site['user'],
@@ -678,7 +894,7 @@ class Hoe
       require 'rubyforge'
       subject, title, body, urls = announcement
 
-      rf = RubyForge.new
+      rf = RubyForge.new.configure
       rf.login
       rf.post_news(rubyforge_name, subject, "#{title}\n\n#{body}")
       puts "Posted to rubyforge"
@@ -693,7 +909,8 @@ class Hoe
       require 'find'
       files = []
       with_config do |config, _|
-        exclusions = config["exclude"] || /tmp$|CVS|\.svn/
+        exclusions = config["exclude"]
+        abort "exclude entry missing from .hoerc. Aborting." if exclusions.nil?
         Find.find '.' do |path|
           next unless File.file? path
           next if path =~ exclusions
@@ -732,7 +949,7 @@ class Hoe
 
         puts "Installed key and certificate."
 
-        rf = RubyForge.new
+        rf = RubyForge.new.configure
         rf.login
 
         cert_package = "#{rubyforge_name}-certificates"
@@ -760,11 +977,10 @@ class Hoe
 
   def announcement # :nodoc:
     changes = self.changes.rdoc_to_markdown
-
     subject = "#{name} #{version} Released"
-    title = "#{name} version #{version} has been released!"
-    body = "#{description}\n\nChanges:\n\n#{changes}".rdoc_to_markdown
-    urls = Array(url).map { |s| "* <#{s.strip.rdoc_to_markdown}>" }.join("\n")
+    title   = "#{name} version #{version} has been released!"
+    body    = "#{description}\n\nChanges:\n\n#{changes}".rdoc_to_markdown
+    urls    = Array(url).map { |s| "* <#{s.strip.rdoc_to_markdown}>" }.join("\n")
 
     return subject, title, body, urls
   end
@@ -774,11 +990,16 @@ class Hoe
     cmd = if test ?f, 'test/test_all.rb' then
             "#{RUBY_FLAGS} test/test_all.rb #{FILTER}"
           else
-            tests = ['test/unit'] + test_globs.map { |g| Dir.glob(g) }.flatten
+            tests = ["rubygems", self.testlib] +
+              test_globs.map { |g| Dir.glob(g) }.flatten
             tests.map! {|f| %Q(require "#{f}")}
             "#{RUBY_FLAGS} -e '#{tests.join("; ")}' #{FILTER}"
           end
+
+    excludes = multiruby_skip.join(":")
+    ENV['EXCLUDED_VERSIONS'] = excludes
     cmd = "multiruby #{cmd}" if multi
+
     send msg, cmd
   end
 
@@ -807,5 +1028,11 @@ end
 
 if $0 == __FILE__ then
   out = `rake -T | egrep -v "redocs|repackage|clobber|trunk"`
-  puts out.gsub(/\#/, '-').gsub(/^rake /, '# * ')
+  if ARGV.empty? then
+    # # default::        Run the default tasks.
+    puts out.gsub(/(\s*)\#/, '::\1').gsub(/^rake /, '# ')
+  else
+    # * default        - Run the default tasks.
+    puts out.gsub(/\#/, '-').gsub(/^rake /, '* ')
+  end
 end
